/**
 * CSignatureLayer - 签名层
 * @author Tevin
 */

<template>
    <AtFloatLayout
        class="c-signature-layer"
        title="手写板 / 请书写签名"
        :isOpened="layerOpened"
        :onClose="evt => handleClose()"
    >
        <view class="c-signature-layer-draw">
            <view class="size-box-top"></view>
            <view class="size-box-bottom"></view>
            <canvas
                class="drawing"
                ref="drawing"
                type="2d"
                :id="cavId"
                :disableScroll="true"
                v-if="cavShow"
                @touchstart="evt => handleWriteStart(evt)"
                @touchmove="evt => handleWriteMove(evt)"
                @touchend="evt => handleWriteEnd(evt)"
            ></canvas>
        </view>
        <view class="c-signature-layer-btns">
            <AtButton
                class="btn-warning"
                type="primary"
                size="large"
                :onClick="evt => handleRestDraw()"
            >重写</AtButton>
            <AtButton
                type="primary"
                size="large"
                :onClick="evt => handleSaveDraw()"
            >完成</AtButton>
        </view>
    </AtFloatLayout>
</template>

<script>
import Taro from '@tarojs/taro';
import { $ } from '@tarojs/extend';
import { AtFloatLayout, AtButton } from 'taro-ui-vue';
import { Tools } from '@components/common/Tools';
import './cSignatureLayer.scss';

export default {
    name: 'CSignatureLayer',
    components: {
        AtFloatLayout,
        AtButton,
    },
    props: {},
    data() {
        return {
            cavId: 'signCanvas-' + Date.now() + '-' + parseInt(Math.random() * 10000),
            layerOpened: false,
            cavShow: false,
            curPoint: {},
            lastPoint: {},
            curLine: [],
            // 第一次触发
            firstTouch: true,
            // 所有笔迹
            chirography: [],
            // 初始画圆的半径
            radius: 1,
            canvasContext: null,
        };
    },
    methods: {
        _initDraw() {
            const query = Taro.createSelectorQuery();
            query
                .select(`#${this.cavId}`)
                .fields({ node: true, size: true })
                .exec(res => {
                    // Canvas 对象
                    const canvas = res[0].node;
                    // Canvas 画布的实际绘制宽高
                    const renderWidth = res[0].width;
                    const renderHeight = res[0].height;
                    // Canvas 绘制上下文
                    this.canvasContext = canvas.getContext('2d');
                    this.canvas = canvas;
                    // 初始化画布大小
                    const dpr = Taro.getSystemInfoSync().pixelRatio;
                    canvas.width = renderWidth * dpr;
                    canvas.height = renderHeight * dpr;
                    this.canvasContext.scale(dpr, dpr);
                    // 初始化变量
                    this.handleRestDraw();
                });
        },
        $onDraw(callback) {
            this.layerOpened = true;
            this._callback = callback;
            this.$nextTick(() => {
                setTimeout(() => {
                    this.cavShow = true;
                    setTimeout(() => {
                        this._initDraw();
                    }, 100);
                }, 300);
            });
        },
        handleRestDraw() {
            this.firstTouch = true;
            this.curLine = [];
            this.chirography = [];
            this.canvasContext.clearRect(0, 0, this.canvas.width, this.canvas.height);
        },
        handleClose() {
            this.layerOpened = false;
            this.cavShow = false;
            setTimeout(() => {
                this.canvasContext.clearRect(0, 0, this.canvas.width, this.canvas.height);
            }, 100);
        },
        handleWriteStart(evt) {
            if (evt.type !== 'touchstart' || !this.canvasContext) {
                return;
            }
            this.canvasContext.fillStyle = '#111111';
            const point = {
                x: evt.touches[0].x,
                y: evt.touches[0].y,
            };
            this.curLine.unshift({
                time: Date.now(),
                dis: 0,
                x: point.x,
                y: point.y,
            });
            if (this.firstTouch) {
                this.cutArea = {
                    top: point.y,
                    right: point.x,
                    bottom: point.y,
                    left: point.x,
                };
                this.firstTouch = false;
            }
            this._pointToLine(this.curLine);
        },
        handleWriteMove(evt) {
            if (evt.type !== 'touchmove') {
                return;
            }
            if (evt.cancelable) {
                // 判断默认行为是否已经被禁用
                if (!evt.defaultPrevented) {
                    evt.preventDefault();
                }
            }
            const point = {
                x: evt.touches[0].x,
                y: evt.touches[0].y,
            };
            this._updateCutArea(point);
            this.lastPoint = this.curPoint;
            this.curPoint = point;
            this.curLine.unshift({
                time: Date.now(),
                dis: this._distance(point, this.lastPoint),
                x: point.x,
                y: point.y,
            });
            this._pointToLine(this.curLine);
        },
        handleWriteEnd(evt) {
            if (evt.type !== 'touchend') {
                return;
            }
            const point = {
                x: evt.changedTouches[0].x,
                y: evt.changedTouches[0].y,
            };
            this.lastPoint = this.curPoint;
            this.curPoint = point;
            this.curLine.unshift({
                time: Date.now(),
                dis: this._distance(point, this.lastPoint),
                x: point.x,
                y: point.y,
            });
            this._pointToLine(this.curLine);
            this.chirography.unshift(this.curLine);
            this.curLine = [];
        },
        // 更新裁剪区域
        _updateCutArea(point) {
            if (point.y < this.cutArea.top) {
                this.cutArea.top = point.y;
            }
            if (point.y < 0) {
                this.cutArea.top = 0;
            }
            if (point.x > this.cutArea.right) {
                this.cutArea.right = point.x;
            }
            if (this.canvas.width - point.x <= 0) {
                this.cutArea.right = this.canvas.width;
            }
            if (point.y > this.cutArea.bottom) {
                this.cutArea.bottom = point.y;
            }
            if (this.canvas.height - point.y <= 0) {
                this.cutArea.bottom = this.canvas.height;
            }
            if (point.x < this.cutArea.left) {
                this.cutArea.left = point.x;
            }
            if (point.x < 0) {
                this.cutArea.left = 0;
            }
        },
        // 求两点之间距离
        _distance(a, b) {
            let x = b.x - a.x;
            let y = b.y - a.y;
            return Math.sqrt(x * x + y * y);
        },
        // 绘制笔迹
        _pointToLine(line) {
            this._calcBethelLine(line);
        },
        // 计算插值的方式
        _calcBethelLine(line) {
            if (line.length <= 1) {
                line[0].r = this.radius;
                return;
            }
            // 笔迹倍数
            const lineSize = 1.5;
            // 最小笔画半径
            const lineMin = 0.5;
            // 最大笔画半径
            const lineMax = 4;
            // 默认压力
            const pressure = 1;
            // 顺滑度，用60的距离来计算速度
            const smoothness = 60;
            // 开始计算
            let x0,
                x1,
                x2,
                y0,
                y1,
                y2,
                r0,
                r1,
                r2,
                len,
                lastRadius,
                dis = 0,
                time = 0,
                curveValue = 0.5;
            if (line.length <= 2) {
                x0 = line[1].x;
                y0 = line[1].y;
                x2 = line[1].x + (line[0].x - line[1].x) * curveValue;
                y2 = line[1].y + (line[0].y - line[1].y) * curveValue;
                //x2 = line[1].x;
                //y2 = line[1].y;
                x1 = x0 + (x2 - x0) * curveValue;
                y1 = y0 + (y2 - y0) * curveValue;
            } else {
                x0 = line[2].x + (line[1].x - line[2].x) * curveValue;
                y0 = line[2].y + (line[1].y - line[2].y) * curveValue;
                x1 = line[1].x;
                y1 = line[1].y;
                x2 = x1 + (line[0].x - x1) * curveValue;
                y2 = y1 + (line[0].y - y1) * curveValue;
            }
            // 从计算公式看，三个点分别是(x0,y0),(x1,y1),(x2,y2) ；(x1,y1)这个是控制点，控制点不会落在曲线上；实际上，这个点还会手写获取的实际点，却落在曲线上
            len = this._distance({ x: x2, y: y2 }, { x: x0, y: y0 });
            lastRadius = this.radius;
            for (let n = 0; n < line.length - 1; n++) {
                dis += line[n].dis;
                time += line[n].time - line[n + 1].time;
                if (dis > smoothness) {
                    break;
                }
            }
            this.radius = Math.min((time / len) * pressure + lineMin, lineMax) * lineSize;
            line[0].r = this.radius;
            //计算笔迹半径；
            if (line.length <= 2) {
                r0 = (lastRadius + this.radius) / 2;
                r1 = r0;
                r2 = r1;
                //return;
            } else {
                r0 = (line[2].r + line[1].r) / 2;
                r1 = line[1].r;
                r2 = (line[1].r + line[0].r) / 2;
            }
            let n = 5;
            let points = [];
            for (let i = 0; i < n; i++) {
                let t = i / (n - 1);
                let x = (1 - t) * (1 - t) * x0 + 2 * t * (1 - t) * x1 + t * t * x2;
                let y = (1 - t) * (1 - t) * y0 + 2 * t * (1 - t) * y1 + t * t * y2;
                let r = lastRadius + ((this.radius - lastRadius) / n) * i;
                points.push({ x: x, y: y, r: r });
                if (points.length == 3) {
                    this._bethelDraw(
                        this._ctaCalc(
                            points[0].x,
                            points[0].y,
                            points[0].r,
                            points[1].x,
                            points[1].y,
                            points[1].r,
                            points[2].x,
                            points[2].y,
                            points[2].r,
                        ),
                    );
                    points = [{ x: x, y: y, r: r }];
                }
            }
            this.curLine = line;
        },
        _ctaCalc(x0, y0, r0, x1, y1, r1, x2, y2, r2) {
            let a = [],
                vx01,
                vy01,
                norm,
                n_x0,
                n_y0,
                vx21,
                vy21,
                n_x2,
                n_y2;
            vx01 = x1 - x0;
            vy01 = y1 - y0;
            norm = Math.sqrt(vx01 * vx01 + vy01 * vy01 + 0.0001) * 2;
            vx01 = (vx01 / norm) * r0;
            vy01 = (vy01 / norm) * r0;
            n_x0 = vy01;
            n_y0 = -vx01;
            vx21 = x1 - x2;
            vy21 = y1 - y2;
            norm = Math.sqrt(vx21 * vx21 + vy21 * vy21 + 0.0001) * 2;
            vx21 = (vx21 / norm) * r2;
            vy21 = (vy21 / norm) * r2;
            n_x2 = -vy21;
            n_y2 = vx21;
            a.push({ mx: x0 + n_x0, my: y0 + n_y0, color: '#1A1A1A' });
            a.push({
                c1x: x1 + n_x0,
                c1y: y1 + n_y0,
                c2x: x1 + n_x2,
                c2y: y1 + n_y2,
                ex: x2 + n_x2,
                ey: y2 + n_y2,
            });
            a.push({
                c1x: x2 + n_x2 - vx21,
                c1y: y2 + n_y2 - vy21,
                c2x: x2 - n_x2 - vx21,
                c2y: y2 - n_y2 - vy21,
                ex: x2 - n_x2,
                ey: y2 - n_y2,
            });
            a.push({
                c1x: x1 - n_x2,
                c1y: y1 - n_y2,
                c2x: x1 - n_x0,
                c2y: y1 - n_y0,
                ex: x0 - n_x0,
                ey: y0 - n_y0,
            });
            a.push({
                c1x: x0 - n_x0 - vx01,
                c1y: y0 - n_y0 - vy01,
                c2x: x0 + n_x0 - vx01,
                c2y: y0 + n_y0 - vy01,
                ex: x0 + n_x0,
                ey: y0 + n_y0,
            });
            a[0].mx = a[0].mx.toFixed(1);
            a[0].mx = parseFloat(a[0].mx);
            a[0].my = a[0].my.toFixed(1);
            a[0].my = parseFloat(a[0].my);
            for (let i = 1; i < a.length; i++) {
                a[i].c1x = a[i].c1x.toFixed(1);
                a[i].c1x = parseFloat(a[i].c1x);
                a[i].c1y = a[i].c1y.toFixed(1);
                a[i].c1y = parseFloat(a[i].c1y);
                a[i].c2x = a[i].c2x.toFixed(1);
                a[i].c2x = parseFloat(a[i].c2x);
                a[i].c2y = a[i].c2y.toFixed(1);
                a[i].c2y = parseFloat(a[i].c2y);
                a[i].ex = a[i].ex.toFixed(1);
                a[i].ex = parseFloat(a[i].ex);
                a[i].ey = a[i].ey.toFixed(1);
                a[i].ey = parseFloat(a[i].ey);
            }
            return a;
        },
        _bethelDraw(points) {
            const ctx = this.canvasContext;
            ctx.beginPath();
            ctx.moveTo(points[0].mx, points[0].my);
            for (let i = 1; i < points.length; i++) {
                ctx.bezierCurveTo(
                    points[i].c1x,
                    points[i].c1y,
                    points[i].c2x,
                    points[i].c2y,
                    points[i].ex,
                    points[i].ey,
                );
            }
            ctx.closePath();
            ctx.stroke();
            ctx.fill();
        },
        _brushingGround(callback) {
            Taro.canvasToTempFilePath({
                canvas: this.canvas,
                x: 0,
                y: 0,
                width: Math.ceil(this.canvas.width),
                height: Math.ceil(this.canvas.height),
                destWidth: Math.ceil(this.canvas.width),
                destHeight: Math.ceil(this.canvas.height),
                quality: 1,
                fileType: 'png',
                success: res => {
                    const ctx = this.canvasContext;
                    ctx.fillStyle = '#ffffff';
                    ctx.fillRect(0, 0, this.canvas.width, this.canvas.height);
                    // 重新绘制
                    const image = this.canvas.createImage();
                    image.onload = () => {
                        const dpr = Taro.getSystemInfoSync().pixelRatio;
                        ctx.drawImage(
                            image,
                            0,
                            0,
                            this.canvas.width / dpr,
                            this.canvas.height / dpr,
                        );
                        callback();
                    };
                    image.src = res.tempFilePath;
                },
            });
        },
        handleSaveDraw() {
            if (this.firstTouch) {
                Tools.toast('请书写签名！');
                return;
            }
            this._brushingGround(() => {
                const delta = 20;
                const clipArea = { x: 0, y: 0, w: 0, h: 0 };
                clipArea.x = Math.max(this.cutArea.left - delta, 0);
                clipArea.y = Math.max(this.cutArea.top - delta, 0);
                const realRight = Math.min(this.cutArea.right + delta, this.canvas.width);
                const realBottom = Math.min(
                    this.cutArea.bottom + delta,
                    this.canvas.height,
                );
                clipArea.w = realRight - clipArea.x;
                clipArea.h = realBottom - clipArea.y;
                Taro.canvasToTempFilePath({
                    canvas: this.canvas,
                    x: clipArea.x,
                    y: clipArea.y,
                    width: clipArea.w,
                    height: clipArea.h,
                    destWidth: clipArea.w,
                    destHeight: clipArea.h,
                    quality: 0.6,
                    fileType: 'jpeg',
                    success: res => {
                        this._callback(res.tempFilePath);
                        this.layerOpened = false;
                    },
                });
            });
        },
    },
    mounted() {},
};
</script>